/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
 * Linux Bandwidth Manager v0.1
 * by Bertrand Goareguer <bertrand.goareguer@gmail.com>
 *
 * to compile : gcc lbm.c -o lbm -Wall
 * to start : sudo ./lbm --input-interface eth0
 *
 * TODO:
 *	- only select the incomming data flows (/sbin/ifconfig | grep eth1 -A 1  | grep "inet adr")
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <signal.h>

#include <gtk/gtk.h>

#include "lbm.h"
#include "configuration_file_creation.h"


unsigned int filter_handles[MAX_FILTER_HANDLES];

//
//create a new connection to add to the list of active connections
//
connection_info_p create_connection(char proto[8],char remote_addr[16],unsigned int remote_port,unsigned int local_port,char state[32],unsigned int pid)
{
	connection_info_p new_connection=NULL;

	new_connection=(connection_info_p)malloc(sizeof(connection_info_t));

	strcpy(new_connection->proto,proto);
	strcpy(new_connection->remote_addr,remote_addr);
	new_connection->remote_port=remote_port;
	new_connection->local_port=local_port;
	strcpy(new_connection->state,state);
	new_connection->pid=pid;
	new_connection->priority=1;
	memset(new_connection->path,0,sizeof(new_connection->path));
	new_connection->rate=-1;
	new_connection->filter_handle=-1;
//	new_connection->manual_priority=0;

	new_connection->next=NULL;

	return new_connection;
}
	 

//
//print a single connection
//
void print_connection(connection_info_p connection)
{
	if(connection!=NULL)
	{
		printf("proto : %s\n",connection->proto);
		printf("remote_addr : %s\n",connection->remote_addr);
		printf("remote_port : %d\n",connection->remote_port);
		printf("local_port : %d\n",connection->local_port);
		printf("state : %s\n",connection->state);
		printf("pid : %d\n",connection->pid);
		printf("process_name : %s\n",connection->process_name);
		printf("priority : %d\n",connection->priority);
		printf("path : %s\n",connection->path);
		printf("rate : %i\n",connection->rate);
		printf("filter handle : %d\n", connection->filter_handle);
//		printf("manual_priority : %d\n",connection->manual_priority);
	}
}

//
//print the list of all active connections
//
void print_connections_list(connection_info_p connections_list)
{
	print_connection(connections_list);
	if(connections_list==NULL)
		return;
	if(connections_list->next!=NULL)
		print_connections_list(connections_list->next);
}


//
//add a pre-created connection to the list of active connections
//
connection_info_p add_connection(connection_info_p connections_list, connection_info_p new_connection)
{
	if(connections_list==NULL)
		return new_connection;
	else
	{
		new_connection->next=connections_list;
		return new_connection;
	}
}


//
//remove a connection from the list of active connections
//
connection_info_p remove_connection(connection_info_p connections_list, connection_info_p connection)
{
	connection_info_p ptr=NULL;

	if(connections_list==NULL)
		return NULL;
	else
	{
		if(	(strcmp(connections_list->proto, connection->proto)==0) &&
			(strcmp(connections_list->remote_addr, connection->remote_addr)==0) &&
			(connections_list->remote_port==connection->remote_port) &&
			(connections_list->local_port==connection->local_port))
		{
			ptr=connections_list->next;
			free(connections_list);
			return ptr;
		}
		else
		{
			connections_list->next=remove_connection(connections_list->next,connection);
			return connections_list;
			
		}
	}
}



//
//find a connection in the list (return NULL if not found)
//
connection_info_p find_connection(connection_info_p connections_list,connection_info_p target)
{
	if(connections_list==NULL)
		return NULL;
	else
	{
		if(	(strcmp(connections_list->proto, target->proto)==0) &&
			(strcmp(connections_list->remote_addr, target->remote_addr)==0) &&
			(connections_list->remote_port==target->remote_port) &&
			(connections_list->local_port==target->local_port))
		{
			return connections_list;
		}
		else
		{
			return find_connection(connections_list->next,target);
		}
	}	
}



//
//find a connection in the list from a path in the unsorted tree model (return NULL if not found)
//
connection_info_p find_connection_from_path(connection_info_p connections_list,char path[8])
{
	if(connections_list==NULL)
		return NULL;
	else
	{
		if(strcmp(connections_list->path, path)==0)
		{
			return connections_list;
		}
		else
		{
			return find_connection_from_path(connections_list->next,path);
		}
	}	
}


//
//return the number of connections in the list
//
int get_nb_connections_in_list(connection_info_p connections_list)
{
	connection_info_p ptr = NULL;
	unsigned int i=0;

	ptr=connections_list;

	while(ptr!=NULL)
	{
		i++;
		ptr=ptr->next;
	}

	return i;
}




//
//retrieve the number of active TCP and UDP connections
//
int get_nb_connections(void)
{
	FILE * fin = NULL;
	char buffer[1024];
	
	unsigned int nb_connections=0;

	char proto[8];
	char token[16];
	int nb=0;

	fin = fopen("/proc/net/sockstat","r");
	if(fin==NULL)
		return -1;

	while(!feof(fin))
	{
		memset(buffer,sizeof(buffer),0);
		fgets(buffer,sizeof(buffer),fin);

		if( (strncmp(buffer,"TCP:",4) == 0) || (strncmp(buffer,"UDP:",4) == 0) )
		{
			sscanf(buffer, "%s %s %d", proto,token,&nb);
			nb_connections=nb_connections+nb;
		}
	}
	
	fclose(fin);

	return nb_connections;
}


//
//get a process name from its PID
//
int get_process_name_from_PID(unsigned int pid, char process_name[64])
{
	FILE * fin = NULL;
	char filename[32];
	char buffer[64];

	memset(filename,sizeof(filename),0);
	memset(buffer,sizeof(buffer),0);

	sprintf(filename,"/proc/%d/cmdline",pid);

	fin=fopen(filename,"r");
	if(fin!=NULL)
	{
		fgets(buffer,sizeof(buffer),fin);
		strcpy(process_name,buffer);	
		fclose(fin);	
		return 0;
	}
	else
	{
		strcpy(process_name,"unknown");
		return (-1);
	}	
	
}



//
//create a connection_info structure for storing a rule
//
connection_info_p create_rule(char * rule, unsigned int *nb_params)
{
	connection_info_p new_rule=NULL;

	char proto[8]="";
	char remote_addr[16]="";
	unsigned int remote_port=0;
	char s_remote_port[8]="";
	unsigned int local_port=0;
	char s_local_port[8]="";
	char process_name[64]="";
	unsigned int priority=0;
	char s_priority[8]="";

	// variables for strtok
	char * saveptr;
	char * token;

	new_rule=(connection_info_p)malloc(sizeof(connection_info_t));

	strcpy(new_rule->state,"");
	new_rule->pid=0;
	strcpy(new_rule->path,"");
	new_rule->rate=0;
	new_rule->filter_handle=0;
	new_rule->next=NULL;


	// initialize the number of parameters to 0	
	(*nb_params)=0;


	// get the proto parameter, if present
	if(strstr(rule,"proto=")!=NULL)
	{
		strncpy(proto,strstr(rule,"proto=")+strlen("proto="),sizeof(proto));

		token=strtok_r(proto," ",&saveptr);
		strcpy(proto,token);
		sscanf(proto,"%s",new_rule->proto);

		(*nb_params)++;
		
	}
	else
		strcpy(new_rule->proto,"\0");



	// get the remote_addr parameter, if present
	if(strstr(rule,"remote_addr=")!=NULL)
	{
		strncpy(remote_addr,strstr(rule,"remote_addr=")+strlen("remote_addr="),sizeof(remote_addr));

		token=strtok_r(remote_addr," ",&saveptr);
		strcpy(remote_addr,token);
		sscanf(remote_addr,"%s",new_rule->remote_addr);

		(*nb_params)++;
		
	}
	else
		strcpy(new_rule->remote_addr,"\0");





	// get the remote_port parameter, if present
	if(strstr(rule,"remote_port=")!=NULL)
	{
		strncpy(s_remote_port,strstr(rule,"remote_port=")+strlen("remote_port="),sizeof(s_remote_port));

		token=strtok_r(s_remote_port," ",&saveptr);
		strcpy(s_remote_port,token);
		sscanf(s_remote_port,"%d",&new_rule->remote_port);

		(*nb_params)++;

		
	}
	else
		new_rule->remote_port=0;



	// get the local_port parameter, if present
	if(strstr(rule,"local_port=")!=NULL)
	{
		strncpy(s_local_port,strstr(rule,"local_port=")+strlen("local_port="),sizeof(s_local_port));

		token=strtok_r(s_local_port," ",&saveptr);
		strcpy(s_local_port,token);
		sscanf(s_local_port,"%d",&new_rule->local_port);

		(*nb_params)++;

		
	}
	else
		new_rule->local_port=0;



	// get the process_name parameter, if present
	if(strstr(rule,"process_name=")!=NULL)
	{
		strncpy(process_name,strstr(rule,"process_name=")+strlen("process_name="),sizeof(process_name));

		token=strtok_r(process_name," ",&saveptr);
		strcpy(process_name,token);
		sscanf(process_name,"%s",new_rule->process_name);

		(*nb_params)++;
		
	}
	else
		strcpy(new_rule->process_name,"\0");




	// get the priority parameter. This parameter MUST be present and MUST NOT be counted as a parameter
	if(strstr(rule,"priority=")!=NULL)
	{
		strncpy(s_priority,strstr(rule,"priority=")+strlen("priority="),sizeof(s_priority));

		token=strtok_r(s_priority," ",&saveptr);
		strcpy(s_priority,token);
		sscanf(s_priority,"%d",&new_rule->priority);
		
	}
	else
		return NULL;

	return new_rule;
}





//
//compute the matching coefficient between a connection ans a rule
//
unsigned int compute_matching_coefficient(connection_info_p connection, connection_info_p rule, unsigned int nb_params)
{
	unsigned int matching_coefficient=0;

	if(strcmp(connection->proto,rule->proto)==0)
		matching_coefficient++;

	if(strcmp(connection->remote_addr,rule->remote_addr)==0)
		matching_coefficient++;

	if(connection->remote_port==rule->remote_port)
		matching_coefficient++;

	if(connection->local_port==rule->local_port)
		matching_coefficient++;

	if(strcmp(connection->process_name,rule->process_name)==0)
		matching_coefficient++;

	if(matching_coefficient == nb_params)
		return matching_coefficient;
	else
		return 0;
}



//
//get the index of the biggest value of a vector
//
int get_max(char *match_vector, int length)
{
	char max_value=match_vector[0];
	int max_index=0;
	unsigned int i=0; // loop counter

	for(i=1;i<length;i++)
	{
		if(match_vector[i]>max_value)
			max_index=i;
		else
		{
			// if 2 rules have the same matching coefficient greater than 0, we can't choose the best
			// then, we return an error code
			if(match_vector[i]==max_value && max_value!=0)
				return -1;
		}
	}

	return max_index;
}



//
//determines the priority order for a given connextion (based on the automatic rules file)
//
unsigned int guess_priority(connection_info_p connection)
{
	char auto_rule_file[128]="";
	char auto_rule_file_new[128]="";
	char auto_rule_file_dir[128]="";
	FILE * fin=NULL;
	FILE * fout=NULL;
	char buffer[1024];
	char buffer_new[1024];
	unsigned int nb_rules=0; // number of rules found in the automatic rules file
	char *match_vector=NULL; // match coefficient between the given connection and each rule
	unsigned int * priority_vector=NULL; // vector containing the priorities for each rule
	unsigned int i=0; // loop counter
	connection_info_p rule=NULL;
	int index=0; // index of the rule which has the biggest matching coefficient for a specific connection
	unsigned int nb_params=0;
	char command[128]="";
	struct stat sts;

	sprintf(auto_rule_file,"%s/.lbm/auto_rules.txt",HOME);
	sprintf(auto_rule_file_new,"%s/.lbm/auto_rules.txt.new",HOME);


	// check if the .lbm directory is present in the user's home directory
	sprintf(auto_rule_file_dir,"%s/.lbm",HOME);
	if (stat(auto_rule_file_dir,&sts) != 0)
	{
		printf ("%s does not exist\n", auto_rule_file_dir);

		// create the directory
		sprintf(command,"mkdir %s/.lbm",HOME);
		system(command);		
	}


	fout=fopen(auto_rule_file_new,"w");
	if(fout==NULL)
		return DEFAULT_PRIORITY_ORDER;

	fin=fopen(auto_rule_file,"r");
	if(fin==NULL)
		return DEFAULT_PRIORITY_ORDER;
	
	// the automatic rules file is present, lets use it !
	
	// find the number of rules in the automatic rules file
	while(!feof(fin))
	{
		memset(buffer,0,sizeof(buffer));
		fgets(buffer,sizeof(buffer),fin);
		if(buffer[0]!='#' && strlen(buffer)>5)
			nb_rules++;
	}

	if(nb_rules==0)
		return DEFAULT_PRIORITY_ORDER; // return the default priority


	// allocate memory for the match vector
	match_vector=(char*)malloc(nb_rules*sizeof(char));

	// allocate memory for the priority vector
	priority_vector=(unsigned int*)malloc(nb_rules*sizeof(unsigned int));
	
	// initialize the match_vector with the all-zero vector
	memset(match_vector,0,nb_rules);

	// initialize the priority vector with the all-one vector
	for(i=0;i<nb_rules;i++)
		priority_vector[i]=1;

	// come back to the beginning of the file
	if(fseek(fin,0,SEEK_SET)!=0)
	{
		fclose(fin);
		return DEFAULT_PRIORITY_ORDER;
	}

	// initialize the current rule index to 0
	i=0;

	// read the file line by line
	while(!feof(fin))
	{
		// get a rule
		memset(buffer,0,sizeof(buffer));
		fgets(buffer,sizeof(buffer),fin);
		if(buffer[0]!='#' && strlen(buffer)>5) // if it's not a comment line or a blank line
		{
			// we need to parse all the elements of the rule
			printf("buffer=%s\n",buffer);
			rule=create_rule(buffer,&nb_params);
			if(rule==NULL)
			{		
				memset(buffer_new,0,sizeof(buffer_new));
				strncpy(buffer_new+2,buffer,sizeof(buffer_new)-2);
				buffer_new[0]='#';
				buffer_new[1]=' ';

				fputs(buffer_new,fout);

				printf("The following rule does not specify a priority order\n\n%s\n",buffer);
				printf("This rule has been ignored\n\n");

				dialog = gtk_message_dialog_new ((GtkWindow*)window,
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_MESSAGE_ERROR,
                                  GTK_BUTTONS_CLOSE,
                                  "The following rule does not specify a priority order\n%s\nThis rule has been ignored",buffer);

				gtk_dialog_run (GTK_DIALOG (dialog));
				gtk_widget_destroy(dialog);
			}
			else
			{
				fputs(buffer,fout);
				print_connection(rule);
				printf("\n");
			}

			// we can now compute the match coefficient between the given connection and the current rule
			// if the current rule is mis-written (e.g. missing priority order), the matching coefficient is 0
			if(rule==NULL)
				match_vector[i]=0;				
			else
			{
				priority_vector[i]=rule->priority;
				match_vector[i]=compute_matching_coefficient(connection, rule, nb_params);
			}	
			
			i++; // go to next rule
		}
		else
		{
			fputs(buffer,fout);
		}
	}

	fclose(fin);
	fclose(fout);

	sprintf(command,"mv -f %s %s",auto_rule_file_new, auto_rule_file);
	system(command);


	// inspect the match vector to find the rule fitting best with the given connection
	index=get_max(match_vector, nb_rules);

	
	if(index<0)
	{
		printf("Warning : Two (or more) automatic rules gave the same matching coefficient. Using the default priority order.\n\n");

		dialog = gtk_message_dialog_new ((GtkWindow*)window,
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_MESSAGE_ERROR,
                                  GTK_BUTTONS_CLOSE,
                                  "Warning : Two (or more) automatic rules gave the same matching coefficient. Using the default priority order.");

		gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy(dialog);
		

		return DEFAULT_PRIORITY_ORDER;
	}
	else
	{
		// if no rule matched the connection, we return the default priority order
		if(match_vector[index]==0)
			return DEFAULT_PRIORITY_ORDER;

		else 	
			return priority_vector[index];
			
	}

	// we should never get there, but just in case
	return DEFAULT_PRIORITY_ORDER;

}


//
//fill the list of all active connections
//
connection_info_p fill_connections_list(connection_info_p connections_list)
{
	FILE * fin = NULL;
	char buffer[1024];

	char proto[8];
	unsigned int recv_Q;
	unsigned int send_Q;
	char local[32];
	char remote[32];
	char state[32];
	unsigned int uid;
	unsigned int inode;
	char s_pid[256];

	char remote_addr[16];
	unsigned int remote_port;
	unsigned int local_port;
	unsigned int pid;

	// variables for strtok
	char * saveptr;
	char * token;

	char * ptr;
	

	connection_info_p new_connection = NULL;

	system("netstat -utapen > ./netstat.txt");

	fin=fopen("./netstat.txt","r");
	if(fin==NULL)
		return NULL;

	// get the 2 first lines
	fgets(buffer, sizeof(buffer), fin);
	fgets(buffer, sizeof(buffer), fin);

/*
	// get the total number of active connections
	nb_connections=get_nb_connections();
	if(nb_connections <= 0)
		return NULL;
*/


	// get all the connections one at a time
	while(!feof(fin))
	{
		memset(buffer,sizeof(buffer),0);
		memset(proto,sizeof(proto),0);
		memset(remote,sizeof(remote),0);
		memset(local,sizeof(local),0);
		memset(state,sizeof(state),0);
		memset(s_pid,sizeof(s_pid),0);
		pid=0;

		fgets(buffer,sizeof(buffer),fin);

		if( (strncmp(buffer,"tcp ",4) == 0) || (strncmp(buffer,"udp ",4) == 0) )
		{
			if((ptr=strstr(buffer,"ESTABLISHED"))!=NULL)
			{
				*(ptr+9)=' ';
				*(ptr+10)=' ';
			

				sscanf(buffer,"%s %d %d %s %s %s %d %d %s",
					proto,
					&recv_Q,
					&send_Q,
					local,
					remote,
					state,
					&uid,
					&inode,
					s_pid);

				// rename the ESTABLISH state
				if(strcmp(state,"ESTABLISH")==0)
				{
					strcpy(state,"ESTABLISHED");
				}

				// separate local addr and local port
				token=strtok_r(local,":",&saveptr);
				//strcpy(local_addr,token);
				if(saveptr[0]=='*')
					local_port=0;
				else
					sscanf(saveptr,"%d",&local_port);

				// separate remote addr and remote port
				token=strtok_r(remote,":",&saveptr);
				strcpy(remote_addr,token);
				if(saveptr[0]=='*')
					remote_port=0;
				else
					sscanf(saveptr,"%d",&remote_port);		

				// get the PID from the PID/process_name field
				token=strtok_r(s_pid,"/",&saveptr);
				sscanf(token,"%d",&pid);

				new_connection=create_connection(proto,remote_addr,remote_port,local_port,state,pid);

				// get the complete process name
				if(get_process_name_from_PID(new_connection->pid,new_connection->process_name) != -1)
					connections_list=add_connection(connections_list,new_connection);
				else
				{
					// if the process name could not be retrieved, it means that the process has terminated
					// so the connection has been closed and we don't need to add it to the list
					free(new_connection);
					new_connection=NULL;
				}

			}
		}

	}	

	fclose(fin);

	system("rm -f netstat.txt");

	return connections_list;

}


//
//affect priorities to each active flow
//
connection_info_p affect_priorities(connection_info_p connections_list)
{
	if(connections_list!=NULL)
	{
		print_connection(connections_list);
		printf("priority : ");
		scanf("%d",&connections_list->priority);

		printf("rate : ");
		scanf("%i",&connections_list->rate);

		connections_list->next=affect_priorities(connections_list->next);
	}

	return connections_list;
}

//
//free the list of all active connections
//
connection_info_p free_connections_list(connection_info_p connections_list)
{
	connection_info_p next = NULL;

	if(connections_list!=NULL)
	{
		if(connections_list->next==NULL)
		{
			free(connections_list);
			return NULL;
		}
		else
		{
			next=connections_list->next;
			free(connections_list);
			connections_list=free_connections_list(next);
		}
	}

	return connections_list;

}


//
//clear qdiscs (clear everything related to TC)
//
void clear_qdiscs(void)
{
	char buffer[1024];

	memset(buffer,0,sizeof(buffer));

	system("modprobe ifb");
	system("ifconfig ifb0 up");

	system("/sbin/tc qdisc del dev ifb0 root");

	sprintf(buffer,"/sbin/tc qdisc del dev %s ingress",input_interface);
	system(buffer);
}


//
//Re-affect priorities to all active connections based on the automatic rules file
//the default priority order is affected to connections matching none of the rules
//
void reload_automatic_rules(void)
{
	connection_info_p ptr=connections_list;
	while(ptr!=NULL)
	{
		ptr->priority=guess_priority(ptr);
		ptr=ptr->next;
	}
	
}



//
//refresh the list of all active connections
//the priorities that have been set by the user have to be kept
//closed connections have ot be removed
//new connections have to be added
//
connection_info_p refresh_connections_list(connection_info_p connections_list)
{
	connection_info_p new_connections_list=NULL;
	connection_info_p ptr=NULL;
	connection_info_p ptr2=NULL;
	

	// find all the active connections at the present time
	new_connections_list=fill_connections_list(new_connections_list);

	//print_connections_list(new_connections_list);

	// for each connection that was already in the old list
	// copy the "priority", "path" and "filter_handle" fields
	ptr=new_connections_list;

	while(ptr!=NULL)
	{
		ptr2=find_connection(connections_list,ptr);
		if( ptr2 != NULL )
		{
			// the filter is already installed, we only have to update the list
			ptr->priority=ptr2->priority;
			strcpy(ptr->path,ptr2->path);
			ptr->filter_handle=ptr2->filter_handle;

			// delete the connection from the old list
			connections_list=remove_connection(connections_list,ptr);
		}
		else
		{
			// this is a new element, so we have to set its priority order
			printf("this is a new element, so we have to set its priority order\n");
			ptr->priority=guess_priority(ptr);
			// the element is up to date in the new list but the corresponding filter hasn't been created yet
			add_filter(ptr);
		}
			
		ptr=ptr->next;
	}

	
	// remove the filters associated to the expired connections
	ptr=connections_list;
	while(ptr!=NULL)
	{
		remove_filter(ptr);
		ptr=ptr->next;
	}

	// copy the new list in the old one
	connections_list=free_connections_list(connections_list);
	
	ptr=new_connections_list;

	while(ptr!=NULL)
	{
		ptr2=create_connection(ptr->proto, ptr->remote_addr, ptr->remote_port, ptr->local_port, ptr->state, ptr->pid);
		if(get_process_name_from_PID(ptr2->pid,ptr2->process_name) != -1)
		{
			ptr2->priority=ptr->priority;
			strcpy(ptr2->path,ptr->path);
			ptr2->filter_handle=ptr->filter_handle;
			connections_list=add_connection(connections_list,ptr2);
		}
		else
		{
			// if the process name could not be retrieved, it means that the process has terminated
			// so the connection has been closed and we don't need to add it to the list
			free(ptr2);
			ptr2=NULL;
		}
		ptr=ptr->next;
	}

	new_connections_list=free_connections_list(new_connections_list);

	return connections_list;
}





//
//build the core structure of the scheduler
//
void build_scheduler(void)
{
	char buffer[1024];

	memset(buffer,0,sizeof(buffer));

	system("modprobe ifb");
	system("ifconfig ifb0 up");
/*
	system("/sbin/tc qdisc del dev ifb0 root");

	sprintf(buffer,"/sbin/tc qdisc del dev %s ingress",input_interface);
	system(buffer);
*/
	clear_qdiscs();

	printf("max_download_bitrate=%d\n",max_download_bitrate);

	system("/sbin/tc qdisc add dev ifb0 root handle 1: htb default 1");
	sprintf(buffer,"/sbin/tc class add dev ifb0 parent 1: classid 1:1 htb rate %dkbit burst 15k", max_download_bitrate);
	system(buffer);
	system("/sbin/tc qdisc add dev ifb0 parent 1:1 handle 2: prio bands 3 priomap 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2");
	sprintf(buffer,"/sbin/tc qdisc add dev ifb0 parent 2:1 handle 10: sfq perturb %d",(int)ceil(REFRESH_TIMEOUT/2));
	system(buffer);
	sprintf(buffer,"/sbin/tc qdisc add dev ifb0 parent 2:2 handle 20: sfq perturb %d",(int)ceil(REFRESH_TIMEOUT/2));
	system(buffer);
	sprintf(buffer,"/sbin/tc qdisc add dev ifb0 parent 2:3 handle 30: sfq perturb %d",(int)ceil(REFRESH_TIMEOUT/2));
	system(buffer);

	sprintf(buffer, "/sbin/tc qdisc add dev %s ingress handle ffff:", input_interface);
	system(buffer);
	sprintf(buffer, "/sbin/tc filter add dev %s parent ffff: protocol ip u32 match u32 0 0 flowid 1:3 action mirred egress redirect dev ifb0", 			input_interface);
	system(buffer);

}


//
//add a single filter
//
void add_filter(connection_info_p connection)
{
	char command[1024];
	int i=0;
	int last_handle=-1;


	while( (filter_handles[i]==1) && (i<MAX_FILTER_HANDLES) )
	{
		i++;
	}
	if(i==MAX_FILTER_HANDLES && filter_handles[i]==1)
		printf("no more available filter handles !\n");
	else
	{
		filter_handles[i]=1;
		last_handle=i;
	}

	// add 1 to last_handle because the filter handle table starts at index 0
	connection->filter_handle=last_handle+1;

	memset(command,sizeof(command),0);

	if(connection->rate==-1)
	{
		sprintf(command,"/sbin/tc filter add dev ifb0 parent 2: handle ::%d protocol ip prio 1 u32 match ip src %s/32 match ip sport %d 0xffff match ip dport %d 0xffff classid 2:%d", connection->filter_handle, connection->remote_addr, connection->remote_port, connection->local_port, connection->priority+1);
	}
	else
	{
		sprintf(command,"/sbin/tc filter add dev ifb0 parent 2: handle ::%d protocol ip prio 1 u32 match ip src %s/32 match ip sport %d 0xffff match ip dport %d 0xffff police rate %ikbit burst 100kbit drop classid 2:%d", connection->filter_handle, connection->remote_addr, connection->remote_port, connection->local_port, connection->rate, connection->priority+1);
	}
	printf("add filter : command=%s\n",command);
	system(command);
	
}



//
//remove a single filter
//
void remove_filter(connection_info_p connection)
{
	char command[1024];

	sprintf(command,"/sbin/tc filter del dev ifb0 parent 2: handle 800::%d protocol ip prio 1 u32 ", connection->filter_handle);

	system(command);
	printf("remove_filter : %s\n",command);

	// make the filter handle available again
	filter_handles[connection->filter_handle-1]=0;

	connection->filter_handle=0;
}



//
//handler function for the SIGINT signal
//
void sigint_handler(int a)
{
	char buffer[1024];

	memset(buffer,0,sizeof(buffer));

	printf("\n\nCleaning up the schedulers before exiting...\n");

	// free memory
	connections_list=free_connections_list(connections_list);

	system("/sbin/tc qdisc del dev ifb0 root");
	sprintf(buffer,"/sbin/tc qdisc del dev %s ingress",input_interface);
	system(buffer);

	exit(0);
}



// on a refresh timer timeout, the list of active connections should be updated
void lbm_refresh_timeout (gpointer data)
{
	printf("refresh timeout\n");

	//print_connections_list(connections_list);

	// fill in the list of all active connections
	//connections_list=fill_connections_list(connections_list);
	connections_list=refresh_connections_list(connections_list);

	print_connections_list(connections_list);

	printf("nb connections in list : %d\n",get_nb_connections_in_list(connections_list));

	//print_connections_list(connections_list);

}

/*
void print_help(void)
{
	printf("usage : ./lbm -i <interface> -b <bitrate> \n\n");
	printf("  -i, --input-interface=INTERFACE \t Specifies the input interface\n");
	printf("  -b, --max-bitrate=BITRATE \t\t Specifies the maximum download rate on the specified interface\n");
	printf("  -h, --help \t\t\t\t Displays this help message\n");
}
*/


//
//function that reads the configuration file to retrieve the input interface name and the max download bandwidth
//
char read_config_file()
{
	FILE * fin=NULL;
	FILE * fout=NULL;
	char buffer[1024];
	char s_max_download_bitrate[1024];

	// variables for strtok
	char * saveptr;
	char * token;

	char * ptr=NULL;

	struct stat sts;
	char config_file[256];
	char config_file_dir[128]="";

	int test;

	gint response;

	// check if the .lbm directory is present in the user's home directory
	sprintf(config_file_dir,"%s/%s",HOME,".lbm");
	if (stat(config_file_dir,&sts) != 0)
	{
		printf ("%s does not exist\n", config_file_dir);

		// create the directory
		char command[128]="";
		sprintf(command,"mkdir %s/.lbm",HOME);
		system(command);		
	}
	
	sprintf(config_file,"%s/lbm.conf",config_file_dir);
	fin=fopen(config_file,"r");
	if(fin==NULL)
	{
		printf("Configuration file does not exist.\n");
		dialog = gtk_message_dialog_new ((GtkWindow*)window,
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_MESSAGE_QUESTION,
                                  GTK_BUTTONS_YES_NO,
                                  "Configuration file does not exist.\nDo you want to create it ?");

		response = gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy(dialog);

		if(response == GTK_RESPONSE_YES)
		{
			create_configuration_file(config_file);
			//printf("response = YES\n");
			
		}


		fin=fopen(config_file,"r");
		if(fin==NULL)
		{
			// user did not want to create the configuration file or cancelled the creation
			printf("Could not find the configuration file.\n");
			dialog = gtk_message_dialog_new ((GtkWindow*)window,
		                          GTK_DIALOG_DESTROY_WITH_PARENT,
		                          GTK_MESSAGE_ERROR,
	                                  GTK_BUTTONS_CLOSE,
		                          "Could not find the configuration file");

			gtk_dialog_run (GTK_DIALOG (dialog));

			gtk_widget_destroy(dialog);
			return -1;
		}
		
		
	}

	// when we get there, the configuration file exists and we can read it
	while(!feof(fin))
	{
		memset(buffer,0,sizeof(buffer));
		fgets(buffer,sizeof(buffer),fin);
		if(buffer[0]!='#' && strlen(buffer)>5)
		{
			if(strncmp(buffer,"input_interface",strlen("input_interface"))==0)
			{
				token=strtok_r(buffer,"=",&saveptr);
				memset(input_interface,0,sizeof(input_interface));
				strncpy(input_interface,saveptr,sizeof(input_interface));
				ptr=strchr(input_interface,'\n');
				if(ptr!=NULL)
					(*ptr)='\0';
			}
			else
			{
				if(strncmp(buffer,"max_download_bitrate",strlen("max_download_bitrate"))==0)
				{
					token=strtok_r(buffer,"=",&saveptr);
					memset(s_max_download_bitrate,0,sizeof(s_max_download_bitrate));
					strncpy(s_max_download_bitrate,saveptr,sizeof(s_max_download_bitrate));
					sscanf(s_max_download_bitrate,"%d",&max_download_bitrate);
				}
			}
		}
	}
	fclose(fin);
	return 0;
	
}

int lbm_main(int argc, char ** argv)
{
	int i=0;

	int c=0;
	int option_index=0;

	char input_interface_set=0;
	char max_download_bitrate_set=0;

	connections_list=NULL;
/*
	if(argc<3 || argc>5)
	{
		print_help();
		exit(0);
	}

	while(c!=-1)
	{
		static struct option long_options[]= {
			{"input-interface",1,0,'i'},
			{"max-download-bitrate",1,0,'b'},
			{"help",0,0,'h'},
			{0,0,0,0}
		};

		c=getopt_long(argc,argv,"i:b:h",long_options,&option_index);

		switch(c)
		{
			case 'i': strcpy(input_interface,optarg);
				  // we should check if the selected interface is really an interface name...
				  input_interface_set=1;
				  break;

			case 'b': max_download_bitrate=atoi(optarg);
				  if(max_download_bitrate>0)
					max_download_bitrate_set=1;
				  break;

			case 'h': print_help();
				  exit(0);

			case '?':
				  break;

			default:  break;
		}
	}

	if(!input_interface_set)
	{
		printf("You have to specify an input network interface\n");
		exit(0);
	}

	if(!max_download_bitrate_set)
	{
		printf("You have to specify a positive maximum download bitrate (in kpbs)\n");
		exit(0);
	}

	printf("input_interface : %s\n",input_interface);
	printf("max_download_bitrate : %d\n",max_download_bitrate);
*/

	if(read_config_file()<0)
		return -1;

	printf("input_interface : %s\n",input_interface);
	printf("max_download_bitrate : %d\n",max_download_bitrate);



	signal(SIGINT,sigint_handler);
	
	printf("Linux Bandwidth Manager\n\n");


	for(i=0;i<MAX_FILTER_HANDLES;i++)
	{
		filter_handles[i]=0;
	}


	
	// build the scheduler
	build_scheduler();

	
	return 0;
}

