/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


/*
 * Linux Bandwidth Manager v0.1
 * by Bertrand Goareguer <bertrand.goareguer@gmail.com>
 *
 * to compile : $ gcc -o lbm `pkg-config --cflags --libs gtk+-2.0` -Wall lbm_ui.c lbm.c
 * to start : sudo ./lbm --input-interface eth0
 *
 * TODO:
 *	- 
 */

/*
 * I used code from Sergio Villar Senn to build the drag-n-drop style GUI
 * This is an example of how to manage drag and drop with a sorted tree model
 *
 * Sergio Villar Senn, 2007. Igalia.
 *
 *  $ gcc -o lbm `pkg-config --cflags --libs gtk+-2.0` -Wall lbm_ui.c lbm.c
 */

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>

#include "lbm.h"



static GtkWidget *my_menu = NULL;
static GtkStatusIcon *status_icon = NULL;




static guint timer_expander;
static guint timer_refresh;

static gint 
cmp_rows (GtkTreeModel *tree_model, 
	  GtkTreeIter *iter1, 
	  GtkTreeIter *iter2,
	  gpointer user_data);

static gint lbm_ui_refresh_timeout (gpointer treeview);



void destroy(GtkWidget *widget,gpointer data)
{
	// free the memory and delete everything we built inside TC
	sigint_handler(0);

	// quit the app
	gtk_main_quit();
}



void update_treestore(GtkWidget *widget,gpointer treeview)
{
	GtkTreeModel * model;
	GtkTreeModel * unsorted_model;
	GtkTreeStore * store;
	GtkTreeIter iter;
	GtkTreeIter *parents;
	GtkTreeModel *model_sort;
	gboolean result=TRUE;
	gchar path[8];
	gchar label[128];
	connection_info_p ptr=NULL;
	GtkTreePath * unsorted_tree_path;

	int i=0;
	


	parents=(GtkTreeIter*)malloc(NB_PRIORITY*sizeof(GtkTreeIter));

	// retrieve the unsorted tree model
	model=gtk_tree_view_get_model(treeview);
	unsorted_model = gtk_tree_model_sort_get_model (GTK_TREE_MODEL_SORT (model));
	store=GTK_TREE_STORE(unsorted_model);


	// detach the model from the tree (speed issues when inserting a lot of rows)
	gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), NULL);

	// remove all children rows
	for(i=0;i<NB_PRIORITY;i++)
	{
		sprintf(path,"%i:0",i);
		if(gtk_tree_model_get_iter(unsorted_model,&iter,gtk_tree_path_new_from_string(path)))
		{
			while(result!=FALSE)
			{
				result=gtk_tree_store_remove(store,&iter);
			}
			result=TRUE;
		}
	}
	


	// insert all children rows from the list of all active connections
	ptr=connections_list;
	while(ptr!=NULL)
	{
		sprintf(path,"%i",ptr->priority);
		if(gtk_tree_model_get_iter(unsorted_model,&parents[ptr->priority],gtk_tree_path_new_from_string(path)))
		{
			// insert all children corresponding to this priority
			gtk_tree_store_insert(store,&iter,&parents[ptr->priority],MAX_CHILDREN_PER_ROOT+1);
			if(ptr->rate==-1)
				sprintf(label,"%s \t %s:%d", ptr->process_name, ptr->remote_addr, ptr->remote_port);
			else
				sprintf(label,"%s \t %s:%d \t rate limit: %i kbps", ptr->process_name, ptr->remote_addr, ptr->remote_port, ptr-> rate);
			gtk_tree_store_set (store, &iter, 0, label, -1);

			// fill in the "path" and "priority" fields of the structure corresponding to this connection
			unsorted_tree_path=gtk_tree_model_get_path(GTK_TREE_MODEL (store),&iter);
			strcpy(ptr->path,gtk_tree_path_to_string(unsorted_tree_path));
			//printf("path : %s\n",gtk_tree_path_to_string(unsorted_tree_path));

			ptr->priority=gtk_tree_path_get_indices(unsorted_tree_path)[0];
		}

		ptr=ptr->next;
	}

	

	// sort the tree
	model_sort = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (store));

	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (model_sort),
					      0, 
					      GTK_SORT_ASCENDING);

	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (model_sort),
					 0,
					 cmp_rows, NULL, NULL);




	// re-attach the sorted model to the treeview
	gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), GTK_TREE_MODEL(model_sort));
	g_object_unref (G_OBJECT (model_sort));


	// expand all the treeview
	gtk_tree_view_expand_all(treeview); 

	free(parents);


	
}


/*
void add_row(GtkWidget *widget,gpointer treeview)
{
	GtkTreeModel * model;
	GtkTreeModel * unsorted_model;
	GtkTreeStore * store;
	GtkTreeIter iter;
	GtkTreeIter parent;
	GtkTreeModel *model_sort;
	gboolean result=TRUE;
	gchar path[8];

	int i=0;
	

	// retrieve the unsorted tree model
	model=gtk_tree_view_get_model(treeview);
	unsorted_model = gtk_tree_model_sort_get_model (GTK_TREE_MODEL_SORT (model));
	store=GTK_TREE_STORE(unsorted_model);


	// detach the model from the tree (speed issues when inserting a lot of rows)
	gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), NULL);

	// remove all children rows
	for(i=0;i<NB_PRIORITY;i++)
	{
		sprintf(path,"%i:0",i);
		if(gtk_tree_model_get_iter(unsorted_model,&iter,gtk_tree_path_new_from_string(path)))
		{
			while(result!=FALSE)
			{
				result=gtk_tree_store_remove(store,&iter);
			}
			result=TRUE;
		}
	}
	


	// insert all children rows
	for(i=0;i<NB_PRIORITY;i++)
	{
		sprintf(path,"%i",i);
		if(gtk_tree_model_get_iter(unsorted_model,&parent,gtk_tree_path_new_from_string(path)))
		{
			// insert all children corresponding to this priority
			gtk_tree_store_insert(store,&iter,&parent,0);
			gtk_tree_store_set (store, &iter, 0, "child1", -1);

			// fill in the "path" and "priority" fields of the structure corresponding to this connection
		}
	}


	

	// sort the tree
	model_sort = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (store));

	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (model_sort),
					      0, 
					      GTK_SORT_ASCENDING);

	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (model_sort),
					 0,
					 cmp_rows, NULL, NULL);




	// re-attach the sorted model to the treeview
	gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), GTK_TREE_MODEL(model_sort));
	g_object_unref (G_OBJECT (model_sort));


	// expand all the treeview
	gtk_tree_view_expand_all(treeview); 


	if(connections_list!=NULL)
		printf("UI : connections_list!=NULL\n");
	else
		printf("UI : connections_list=NULL\n");



	
}
*/



void fct_box_button(GtkWidget *box_button, GtkTreeView *treeview){
	//creation de la zone en bas de la 
	//fenetre pour placer les bouton aide et quitter
	GtkWidget *box;
	GtkWidget *button;

	box=gtk_vbox_new(FALSE,3);
	gtk_box_pack_start(GTK_BOX(box_button),box,TRUE,TRUE,0);

	button=gtk_button_new_with_label("Edit default rules");               //creation des boutons aides et quitter
//	button=gtk_button_new_with_label("Add row");
	gtk_box_pack_start(GTK_BOX(box),button,TRUE,TRUE,0);
//	g_signal_connect(G_OBJECT(button),"clicked",G_CALLBACK(TODO),NULL);
	gtk_widget_show(button);
	
	button=gtk_button_new_with_label("Save current rules");
	gtk_box_pack_start(GTK_BOX(box),button,TRUE,TRUE,0);
//	g_signal_connect(G_OBJECT(button),"clicked",G_CALLBACK(TODO),NULL);
	gtk_widget_show(button);

	button=gtk_button_new_with_label("Reload automatic rules");
	gtk_box_pack_start(GTK_BOX(box),button,TRUE,TRUE,0);
	g_signal_connect(G_OBJECT(button),"clicked",G_CALLBACK(reload_automatic_rules),NULL);
	gtk_widget_show(button);	
	gtk_widget_show(box);


	box=gtk_vbox_new(FALSE,3);
  	button=gtk_button_new_with_label("Quit");
	gtk_box_pack_start(GTK_BOX(box),button,TRUE,TRUE,0);
	g_signal_connect(G_OBJECT(button),"clicked",G_CALLBACK(destroy),NULL);
	gtk_widget_show(button);	
	gtk_box_pack_start(GTK_BOX(box_button),box,TRUE,TRUE,0);
	gtk_widget_show(box);
}



static GtkWidget *
create_window (GtkTreeView *treeview)
{
	GtkWidget *window;
	GtkWidget *scrolledwindow;
	GtkWidget *box_button;
	GtkWidget *box_general;

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (window), "Linux Bandwidth Manager");
	gtk_window_set_default_size (GTK_WINDOW (window), 400, 250);

	scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow),
                                      GTK_POLICY_AUTOMATIC,
                                      GTK_POLICY_AUTOMATIC);

	
	box_general=gtk_vbox_new(FALSE,5);

	


	gtk_box_pack_start(GTK_BOX(box_general),scrolledwindow,TRUE,TRUE,TRUE);

	gtk_container_add (GTK_CONTAINER (scrolledwindow), GTK_WIDGET(treeview));
    	
	box_button=gtk_hbox_new(FALSE,5);
    	gtk_box_pack_start(GTK_BOX(box_general),box_button,FALSE,FALSE,TRUE);
	fct_box_button(box_button, treeview);      


	gtk_container_add (GTK_CONTAINER (window), box_general);


	g_signal_connect ((gpointer) window, "delete_event",
			  //G_CALLBACK (gtk_main_quit), NULL);
			G_CALLBACK (destroy), NULL);

	return window;
}


/*
static GtkWidget *
create_window (GtkTreeView *treeview)
{
	GtkWidget *window;
	GtkWidget *scrolledwindow;
	GtkWidget *box_button;
	GtkWidget *box_general;

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (window), "GtkTreeSortedModel dnd sample");
	gtk_window_set_default_size (GTK_WINDOW (window), 400, 250);

	scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow),
                                      GTK_POLICY_AUTOMATIC,
                                      GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (window), scrolledwindow);

	box_general=gtk_vbox_new(FALSE,5);
    	gtk_container_add(GTK_CONTAINER(scrolledwindow),box_general);

	gtk_container_add (GTK_CONTAINER (box_general), GTK_WIDGET (treeview));

	box_button=gtk_hbox_new(FALSE,5);
    	gtk_box_pack_start(GTK_BOX(box_general),box_button,TRUE,TRUE,TRUE);
	fct_box_button(box_button, treeview);      



	g_signal_connect ((gpointer) window, "delete_event",
			  G_CALLBACK (gtk_main_quit), NULL);

	return window;
}
*/

/*
static GtkWidget *
create_window (GtkTreeView *treeview)
{
	GtkWidget *window;
	GtkWidget *scrolledwindow;

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (window), "GtkTreeSortedModel dnd sample");
	gtk_window_set_default_size (GTK_WINDOW (window), 400, 250);

	scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow),
                                      GTK_POLICY_AUTOMATIC,
                                      GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (window), scrolledwindow);

	gtk_container_add (GTK_CONTAINER (scrolledwindow), GTK_WIDGET (treeview));

	g_signal_connect ((gpointer) window, "delete_event",
			  G_CALLBACK (gtk_main_quit), NULL);

	return window;
}
*/

static gint
cmp_rows (GtkTreeModel *tree_model, 
	  GtkTreeIter *iter1, 
	  GtkTreeIter *iter2,
	  gpointer user_data)
{
	gchar *name1, *name2;
	
	gtk_tree_model_get (tree_model, iter1, 0, &name1, -1);
	gtk_tree_model_get (tree_model, iter2, 0, &name2, -1);

	if (name1 == NULL && name2 == NULL)
		return 0;
	else if (name1 == NULL)
		return 1;
	else if (name2 == NULL)
		return -1;
	else
		return g_ascii_strcasecmp (name1, name2);
}

static GtkTreeModel *
create_model (void)
{
	GtkTreeStore *model;
	GtkTreeModel *model_sort;
	GtkTreeIter iter;
	unsigned int i=0;
	char label[64];

	model = gtk_tree_store_new (1, G_TYPE_STRING);

	// create a base model with NB_PRIORITY root nodes
	// no children nodes are created at this point
	for(i=0;i<NB_PRIORITY;i++)
	{
		gtk_tree_store_append (model, &iter, NULL);
		sprintf(label,"Priority level %i",i);
		gtk_tree_store_set (model, &iter, 0, label, -1);
	}

	

	model_sort = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (model));


	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (model_sort),
					      0, 
					      GTK_SORT_ASCENDING);

	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (model_sort),
					 0,
					 cmp_rows, NULL, NULL);

	return model_sort;
}

static void
add_columns (GtkTreeView * treeview)
{
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;

	column = gtk_tree_view_column_new ();
	//gtk_tree_view_column_set_title (column, "");

	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_column_pack_start (column, renderer, TRUE);
	gtk_tree_view_column_set_attributes (column, renderer,
                         "text", 0, NULL);

	gtk_tree_view_append_column (treeview, column);
}

static void
drag_data_get_cb (GtkWidget *widget, 
		  GdkDragContext *context, 
		  GtkSelectionData *selection_data, 
		  guint info, 
		  guint time, 
		  gpointer data)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model_sort, *model;
	GtkTreeIter iter;
	GtkTreePath *source_row_sort;
	GtkTreePath *source_row;

	printf("dans get_cb\n");
	
	/* Get the path to the selected row */
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (widget));
	gtk_tree_selection_get_selected (selection, &model_sort, &iter);
	source_row_sort = gtk_tree_model_get_path (model_sort, &iter);

	/* Get the unsorted model and the path to the selected row in
	   the unsorted model */
	model = gtk_tree_model_sort_get_model (GTK_TREE_MODEL_SORT (model_sort));
	source_row = gtk_tree_model_sort_convert_path_to_child_path (GTK_TREE_MODEL_SORT (model_sort),
								     source_row_sort);

	// prevent top level rows (depth=1) to be reorganized
	// i.e. only rows with depth>1 can be reorganized
	if(gtk_tree_path_get_depth(source_row)>1)
	{	
		/* Set the row as the drag data */
		gtk_tree_set_row_drag_data (selection_data,
					    model,
					    source_row);

		gtk_tree_path_free (source_row_sort);
		gtk_tree_path_free (source_row);

		// the drag and drop action is not yet terminated
		// the refresh timer will be restarted later on

	}
	else
	{
		// since the drag will stop here, the refresh timer will not be restarted later on
		// therefore we need to restart it here
		// restart the refresh timer
		timer_refresh = g_timeout_add (REFRESH_TIMEOUT, lbm_ui_refresh_timeout, GTK_TREE_VIEW(widget));
	}

	
}

static void 
drag_data_received_cb (GtkWidget *widget, 
		       GdkDragContext *context, 
		       gint x, 
		       gint y, 
		       GtkSelectionData *selection_data, 
		       guint info, 
		       guint time, 
		       gpointer data)
{
	GtkTreeModel *model_sort, *model;
	GtkTreeRowReference *source_row_reference;
 	GtkTreePath *source_row, *dest_row, *child_dest_row;
	GtkTreeViewDropPosition pos;
	gboolean success = FALSE;
	connection_info_p ptr = NULL;


	printf("dans data_received_cb\n");

	/* Do not allow further process */
	g_signal_stop_emission_by_name (widget, "drag-data-received");

	if (selection_data->length == 0)
		goto out;

	/* Get the unsorted model, and the path to the source row */
	model_sort = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));
	gtk_tree_get_row_drag_data (selection_data,
				    &model,
				    &source_row);

	if( (ptr=find_connection_from_path(connections_list,gtk_tree_path_to_string(source_row))) == NULL )
	{
		printf("error : cannot lookup connection in list\n");
		goto out;
	}
	else
	{
		printf("connection successfully found in list\n");
		print_connection(ptr);
	}

	/* Get the path to the destination row. Can not call
	   gtk_tree_view_get_drag_dest_row() because the source row
	   it's not selected anymore */
	gtk_tree_view_get_dest_row_at_pos (GTK_TREE_VIEW (widget),
					   x, y,
					   &dest_row,
					   &pos);

	/* Only allow drops IN other rows */
	if (!dest_row)
		goto out;


	// replace the selected destination row by its top level row (so that where ever the row is dropped it gets moved just under a top level row)
	gint top_level_dest_row_index = gtk_tree_path_get_indices(dest_row)[0];
	GtkTreePath * top_level_dest_row=gtk_tree_path_new_from_indices(top_level_dest_row_index, -1);
	

	/* Get the destination row in the usorted model */
	child_dest_row = 
		gtk_tree_model_sort_convert_path_to_child_path (GTK_TREE_MODEL_SORT (model_sort),
							top_level_dest_row);

	



	gtk_tree_path_free (dest_row);
	gtk_tree_path_free (top_level_dest_row);

	/* Check if the drag is possible */
	if (!gtk_tree_drag_dest_row_drop_possible (GTK_TREE_DRAG_DEST (model),
						   child_dest_row,
						   selection_data)) {
		gtk_tree_path_free (child_dest_row);
		goto out;
	}

	/* Get a row reference to the source path because the path
	   could change after the insertion. The gtk_drag_finish() is
	   not able to delete the source because that, so we have to
	   do it manually */
	source_row_reference = gtk_tree_row_reference_new (model, source_row);
	gtk_tree_path_free (source_row);

	/* Insert the dragged row as a child of the dest row */
	gtk_tree_path_down (child_dest_row);

	//update the connection in the list with the new path and the new priority
	ptr->priority=gtk_tree_path_get_indices(child_dest_row)[0];
	strcpy(ptr->path,gtk_tree_path_to_string(child_dest_row));
	print_connection(ptr);

	// delete and re-create the filter
	remove_filter(ptr);
	add_filter(ptr);

	if (gtk_tree_drag_dest_drag_data_received (GTK_TREE_DRAG_DEST (model),
						   child_dest_row,
						   selection_data)) {

		/* Clean dest row */
		gtk_tree_view_set_drag_dest_row (GTK_TREE_VIEW (widget),
						 NULL,
						 GTK_TREE_VIEW_DROP_BEFORE);

		/* Get the new path of the source row */
		source_row = gtk_tree_row_reference_get_path (source_row_reference);

		/* Delete the source row */
		gtk_tree_drag_source_drag_data_delete (GTK_TREE_DRAG_SOURCE (model),
						       source_row);

		success = TRUE;

		gtk_tree_path_free (source_row);
	}

	gtk_tree_row_reference_free (source_row_reference);
	gtk_tree_path_free (child_dest_row);
 out:

	/* Never delete the source, we do it manually */
	gtk_drag_finish (context, success, FALSE, time);

	
	// restart the refresh timer
	timer_refresh = g_timeout_add (REFRESH_TIMEOUT, lbm_ui_refresh_timeout, GTK_TREE_VIEW(widget));

}

/* This function expands a tree node candidate to be the destination
   row of a drag if it's not already expanded. Comes from the
   gtktreeview.c source */
static gint
expand_row_timeout (gpointer data)
{
  GtkTreeView *tree_view = data;
  GtkTreePath *dest_path = NULL;
  GtkTreeViewDropPosition pos;
  gboolean result = FALSE;

  GDK_THREADS_ENTER ();

  gtk_tree_view_get_drag_dest_row (tree_view,
                                   &dest_path,
                                   &pos);

  if (dest_path &&
      (pos == GTK_TREE_VIEW_DROP_INTO_OR_AFTER ||
       pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)) {
	  gtk_tree_view_expand_row (tree_view, dest_path, FALSE);
	  gtk_tree_path_free (dest_path);
  }
  else {
	  if (dest_path)
		  gtk_tree_path_free (dest_path);
	  
	  result = TRUE;
  }
  
  GDK_THREADS_LEAVE ();

  return result;
}


static gboolean
drag_motion_cb (GtkWidget      *widget,
		GdkDragContext *context,
		gint            x,
		gint            y,
		guint           time,
		gpointer        user_data)  
{
	GtkTreeViewDropPosition pos;
	GtkTreePath *dest_row;

	
	/* Remove all pending expander timeouts */
	if (timer_expander != 0) {
		g_source_remove (timer_expander);
		timer_expander = 0;
	}

	/* Get the destination row */
	gtk_tree_view_get_dest_row_at_pos (GTK_TREE_VIEW (widget),
					   x, y,
					   &dest_row,
					   &pos);

	if (!dest_row)
		return FALSE;

	/* Expand the selected row after 1/2 second */
	if (!gtk_tree_view_row_expanded (GTK_TREE_VIEW (widget), dest_row)) {
		gtk_tree_view_set_drag_dest_row (GTK_TREE_VIEW (widget), dest_row, pos);
		timer_expander = g_timeout_add (500, expand_row_timeout, widget);
	}
	gtk_tree_path_free (dest_row);

	return TRUE;
}


// when a drag motion starts, the tree model should not be updated
static void drag_begin_cb (GtkWidget      *widget,
		GdkDragContext *context,
		gint            x,
		gint            y,
		guint           time,
		gpointer        user_data) 
{
	/* Remove all pending refresh timeouts */
	if (timer_refresh != 0) {
		g_source_remove (timer_refresh);
		timer_refresh = 0;
	}
}
 

/* The types of the data that could be dragged */
static const GtkTargetEntry drag_types[] =
{
	{ "GTK_TREE_MODEL_ROW", GTK_TARGET_SAME_WIDGET, 0 }
};

static void
setup_drag_and_drop (GtkTreeView *self)
{
	/* We could not use the specific treeview dnd API because that
	   does not give us enough control over the process. Most of
	   these functions do not work with sorted model so we have to
	   redefine several behaviours */

	/* Config the treeview as destination for dnd */
	gtk_drag_dest_set (GTK_WIDGET (self),
			   GTK_DEST_DEFAULT_ALL,
			   drag_types,
			   G_N_ELEMENTS (drag_types),
			   GDK_ACTION_MOVE);
	
	gtk_signal_connect(GTK_OBJECT (self),
			   "drag_data_received",
			   GTK_SIGNAL_FUNC(drag_data_received_cb),
			   NULL);
	

	/* Config the treeview as source for dnd */
	gtk_drag_source_set (GTK_WIDGET (self),
			     GDK_BUTTON1_MASK | GDK_BUTTON3_MASK,
			     drag_types,
			     G_N_ELEMENTS (drag_types),
			     GDK_ACTION_MOVE);

	gtk_signal_connect(GTK_OBJECT (self),
			   "drag_begin",
			   GTK_SIGNAL_FUNC(drag_begin_cb),
			   NULL);

	gtk_signal_connect(GTK_OBJECT (self),
			   "drag_motion",
			   GTK_SIGNAL_FUNC(drag_motion_cb),
			   NULL);


	gtk_signal_connect(GTK_OBJECT (self),
			   "drag_data_get",
			   GTK_SIGNAL_FUNC(drag_data_get_cb),
			   NULL);
}


static gint lbm_ui_refresh_timeout (gpointer treeview)
{

	lbm_refresh_timeout (NULL);
	update_treestore(NULL,treeview);

	return TRUE;	
}




static void
activate (GtkStatusIcon* status_icon,
		gpointer user_data)
{
	GtkWidget *p_window = user_data;
 
	g_return_if_fail (p_window != NULL);
  	g_return_if_fail (GTK_IS_WIDGET (p_window));
	 
	if (GTK_WIDGET_VISIBLE (p_window))
	{
		gtk_widget_hide (p_window);
	}
	else
	{
		gtk_widget_show (p_window);
	}

}

static void 
popup(GtkStatusIcon *status_icon,
		guint button,
		guint activate_time,
		gpointer user_data)
{
	//g_debug("'popup-menu' signal triggered");

	if (!my_menu)
	{
		GtkWidget *item;
		my_menu = gtk_menu_new();
/*
		item = gtk_menu_item_new_with_label("Let's blink!");
		gtk_menu_append(my_menu, item);

		g_signal_connect(G_OBJECT(item), "activate",
				G_CALLBACK(on_blink_change),
				GUINT_TO_POINTER(TRUE));


		item = gtk_menu_item_new_with_label("Let's stop blinking!");
		gtk_menu_append(my_menu, item);

		g_signal_connect (G_OBJECT(item), "activate",
				G_CALLBACK(on_blink_change), 
				GUINT_TO_POINTER(FALSE));
*/
		
		item = gtk_menu_item_new_with_label("Quit");

		gtk_menu_append(my_menu, item);
		g_signal_connect (G_OBJECT(item), "activate",
				G_CALLBACK(destroy), 
				NULL);

	}
	
	gtk_widget_show_all(my_menu);

	gtk_menu_popup(GTK_MENU(my_menu),
			NULL,
			NULL,
			gtk_status_icon_position_menu,
			status_icon,
			button,
			activate_time);
}





int main (int argc, char *argv[])
{
	GtkWidget *treeview;
	GtkTreeModel *model;

	char icon_file[256];

	gtk_init (&argc, &argv);


	/* Remove all pending refresh timeouts */
	if (timer_refresh != 0) {
		g_source_remove (timer_refresh);
		timer_refresh = 0;
	}


	
	if(lbm_main(argc,argv)<0)
		return;


	/* Create treeview and model */
	treeview = gtk_tree_view_new ();
	add_columns (GTK_TREE_VIEW (treeview));
	setup_drag_and_drop (GTK_TREE_VIEW (treeview));

	model = create_model ();
	gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), model);
	g_object_unref (G_OBJECT (model));


	/* Create window */
	window = create_window (GTK_TREE_VIEW (treeview));
	gtk_widget_show_all (window);

	// create the icon in the norification area
	sprintf(icon_file,"%s/.lbm/lbm.png",HOME);
	status_icon = gtk_status_icon_new_from_file(icon_file);
	gtk_status_icon_set_visible(status_icon, 1);

	gtk_status_icon_set_tooltip(status_icon, "Linux Bandwidth Manager");


	// connect signals for the notification are icon
	g_signal_connect (G_OBJECT (status_icon), "popup-menu", G_CALLBACK (popup), NULL);
	g_signal_connect (G_OBJECT (status_icon), "activate", G_CALLBACK (activate), window);




	lbm_refresh_timeout (NULL);
	update_treestore(NULL,treeview);


	// start the refresh timer
	timer_refresh = g_timeout_add (REFRESH_TIMEOUT, lbm_ui_refresh_timeout, treeview);

	gtk_main ();
	return 0;
}
